import negspy.coordinates as nc
import numpy as np
import math

class ArgumentLengthDoesNotMatchCellIdsException(Exception):
  pass

class NodeNotFoundException(Exception):
  pass

class Cells:

  """
  Generic Cells class for constructing the json needed for client side rendering of cell segmentations/scatterplots (UMAP, PCA etc.).

  :param json The json resulting from various calls to add_mapping, add_polygon_outline etc. that can be served to the client.
  """

  def __init__(self, cell_ids=[]):
    """
    Constructor method

    :param list cell_ids: A list of cell ids to be shown in Vitessce.  The order of these will be used to determine the order of future additions to this class, like segmentations and scatterplot coordinates.
    """
    self._cell_ids = cell_ids
    self.json = dict(zip(cell_ids, [{} for _ in cell_ids]))

  def add_mapping(self, name, coords):
    """
    Add a (dimensionality reduction) scatterplot mapping to each cell.

    :param str name: The unique identifier for the mapping, like UMAP, tSNE or PCA.
    :param list coords: A list of lists like [[1, 2], [3, 4], ...] in the order of cell_ids for each cell to be mapped to a scatterplot coorindate.
    """
    if len(coords) != len(self._cell_ids):
      raise ArgumentLengthDoesNotMatchCellIdsException('Coordinates length does not match Cell IDs Length')
    if type(name) != str:
      raise TypeError('name argument needs to be a string for adding a scatterplot mapping')
    for idx, id in enumerate(self._cell_ids):
      if 'mappings' not in self.json[id]:
        self.json[id]['mappings'] = { name: coords[idx] }
      else:
        self.json[id]['mappings'][name] = coords[idx]
  
  def add_centroids(self, centroids):
    """
    Add a centroid for a spatial segmentation outline to each cell.

    :param list centroids: A list of lists like [[1, 2], [3, 4], ...] in the order of cell_ids for each cell to be mapped to a centroid coorindate.
    """
    if len(centroids) != len(self._cell_ids):
      raise ArgumentLengthDoesNotMatchCellIdsException('Centroid length does not match Cell IDs Length')
    if type(centroids) != list or any([len(centroid) != 2 or type(centroid) != list for centroid in centroids]):
        raise TypeError('Centroids should be a list of two element lists')
    for idx, id in enumerate(self._cell_ids):
      self.json[id]['xy'] = centroids[idx]


  def add_polygon_outline(self, polygon_outline):
    """
    Add a polygon for a spatial segmentation outline to each cell.

    :param list polygon_outline: A list of lists of lists like [[[1, 2], [3, 4], [5, 6]], [[7, 8], [9, 10], [11, 12]]...] in the order of cell_ids for each cell to be mapped to its segmentation.
    """
    if len(polygon_outline) != len(self._cell_ids):
      raise ArgumentLengthDoesNotMatchCellIdsException('Segmentations length does not match Cell IDs Length')
    for idx, id in enumerate(self._cell_ids):
      if type(polygon_outline[idx]) != list or any([len(coord) != 2 or type(coord) != list for coord in polygon_outline[idx]]):
        raise TypeError(f'Polygon outline for {id} should be a list of two element lists i.e xy coordinates')
      self.json[id]['poly'] = polygon_outline[idx]

class CellSets:

  """
  Generic CellSets class for constructing the json needed for client side rendering of the cell sets.

  :param json The json resulting from various calls to add_node that can be served to the client.
  """


  def __init__(self):
    """
    Constructor method
    """

    self.json = {
        "datatype": "cell",
        "version": "0.1.2",
        "tree": []
    }
  
  def add_level_zero_node(self, name):
    """
    Add a new level zero node to the root of the tree.

    :param str name: Name for the new node
    """
    self.json['tree'].append({
        "name": name,
        "children": []
    })

  def add_node(self, name, parent_path, cell_set=None):
    """
    Add a node to a parent node.

    :param str name: Name for the new node
    :param list parent_path: List of strings representing the internal nodes to traverse to reach the desired parent node to which we will add the new node, like ['epithelial', 'meso-epithelial']
    :param list cell_set: List of cell ids which will be added to the new node as part of the set.
    """
    parent_node = self._tree_find_node_by_path(parent_path)
    if parent_node is None:
      raise NodeNotFoundException(f'No node with path {parent_path} found to add {name} to')
    new_node = { "name": name }
    if cell_set:
      new_node['set'] = cell_set
    if 'children' not in parent_node:
      parent_node['children'] = [new_node]
    else:
      parent_node['children'].append(new_node)
  
  def _find_node_by_path(self, node, path, curr_index):
    curr_node_name = path[curr_index]
    if node['name'] == curr_node_name:
      if curr_index == len(path) - 1:
        return node
      if 'children' in node:
        found_nodes = [
          self._find_node_by_path(child, path, curr_index + 1) for child in node['children']
        ]
        found_nodes_not_none = [n for n in found_nodes if n]
        if len(found_nodes_not_none) == 1:
          return found_nodes[0]
    return None

  def _tree_find_node_by_path(self, path):
    found_nodes = [self._find_node_by_path(node, path, 0) for node in self.json['tree']]
    found_nodes_not_none = [n for n in found_nodes if n]
    if len(found_nodes_not_none) == 1:
      return found_nodes_not_none[0]
    return None

class Molecules():

  """
  Generic Molecules class for constructing the json needed for client side rendering of spot data.

  :param json The json resulting from various calls to add_molecule.
  """

  def __init__(self):
    """
    Constructor method
    """
    self.json = {}

  def add_molecule(self, name, coords):
    """
    Add a molecules to a parent node.

    :param str name: Name for the new molecules
    :param list coords: A list of lists like [[1, 2], [3, 4], ...] or [[1, 2, 3], [3, 4, 5], ...] which denote where in xy space the spot data should be placed for the desired name.
    """
    self.json[name] = coords

class GenomicProfiles():
  
  """
  Generic class for representing genomic profiles.
  """

  def __init__(self, f, profile_paths, assembly='hg38', starting_resolution=5000, name="Genomic Profiles"):
    """
    Constructor method

    :param f: The opened Zarr store object.
    :type f: zarr.Group
    :param list[list[str]] profile_paths: A list of cell set paths, one path for each profile.
    :param str assembly: The genome assembly to use for chromosome lengths, passed to negspy. By default, 'hg38'.
    :param int starting_resolution: The starting resolution. By default, 5000.
    :param str name: The name for this set of profiles. By default, 'Genomic Profiles'.
    """

    self.f = f

    num_profiles = len(profile_paths)

    compressor = 'default'

    chromosomes = [ str(chr_name) for chr_name in nc.get_chromorder(assembly)[:25] ] # TODO: should more than chr1-chrM be used?
    num_chromosomes = len(chromosomes)
    chroms_length_arr = np.array([ nc.get_chrominfo(assembly).chrom_lengths[x] for x in chromosomes ], dtype="i8")
    chroms_cumsum_arr = np.concatenate((np.array([0]), np.cumsum(chroms_length_arr)))

    chromosomes_set = set(chromosomes)
    chrom_name_to_length = dict(zip(chromosomes, chroms_length_arr))
    chrom_name_to_cumsum = dict(zip(chromosomes, chroms_cumsum_arr))

    # Prepare to fill in resolutions datasets.
    resolutions = [ starting_resolution*(2**x) for x in range(16) ]

    chromosomes_group = f.create_group("chromosomes")
    for chr_name, chr_len in chrom_name_to_length.items():
      chr_group = chromosomes_group.create_group(chr_name)
      # Create each resolution group.
      for resolution in resolutions:
        chr_shape = (num_profiles, math.ceil(chr_len / resolution))
        chr_group.create_dataset(str(resolution), shape=chr_shape, dtype="f4", fill_value=np.nan, compressor=compressor)

    # f.attrs should contain the properties required for HiGlass's "tileset_info" requests.
    f.attrs['row_infos'] = [
        { "path": profile_path }
        for profile_path in profile_paths
    ]
    f.attrs['resolutions'] = sorted(resolutions, reverse=True)
    f.attrs['shape'] = [ num_profiles, 256 ]
    f.attrs['name'] = name
    f.attrs['coordSystem'] = assembly

    self.resolutions = resolutions
    self.chromosomes = chromosomes
    self.chromosomes_group = chromosomes_group
    self.chrom_name_to_length = chrom_name_to_length
    self.num_profiles = num_profiles
    
    # https://github.com/zarr-developers/zarr-specs/issues/50
    f.attrs['multiscales'] = [
        {
            "version": "0.1",
            "name": chr_name,
            "datasets": [
                { "path": f"chromosomes/{chr_name}/{resolution}" }
                for resolution in sorted(resolutions, reverse=True)
            ],
            "type": "zarr-multivec",
            "metadata": {
                "chromoffset": int(chrom_name_to_cumsum[chr_name]),
                "chromsize": int(chr_len),
            }
        }
        for (chr_name, chr_len) in list(zip(chromosomes, chroms_length_arr))
    ]
  
  def add_profile(self, values, chr_name, profile_index):
    """
    Add a single genomic profile to the output store. This function will aggregate for each resolution.
    
    :param values: A profile array for one chromosome.
    :type values: np.array
    :param str chr_name: The name the chromosome corresponding to this array.
    :param int profile_index: The index of this profile among the list of profiles.
    """
    chromosomes_group = self.chromosomes_group
    num_profiles = self.num_profiles
    resolutions = self.resolutions
    resolution_exps = [ (2**x) for x in range(len(resolutions)) ]

    chr_len = self.chrom_name_to_length[chr_name]
    # Fill in the data for this cluster and chromosome at each resolution.
    for resolution, resolution_exp in zip(resolutions, resolution_exps):
      arr_len = math.ceil(chr_len / resolution)
      chr_shape = (num_profiles, arr_len)

      # Pad the array of values with zeros if necessary before reshaping.
      padding_len = resolution_exp - (values.shape[0] % resolution_exp)
      if values.shape[0] % resolution_exp > 0:
        values = np.concatenate((values, np.zeros((padding_len,))))
      num_tiles = chr_shape[1]
      # Reshape to be able to sum every `resolution_exp` number of values.
      arr = np.reshape(values, (-1, resolution_exp)).sum(axis=-1)

      padding_len = arr_len - arr.shape[0]
      if padding_len > 0:
        arr = np.concatenate((arr, np.zeros((padding_len,))))
      # Set the array in the Zarr store.
      chromosomes_group[chr_name][str(resolution)][profile_index,:] = arr


import json
import unittest

from vitessce import (
    VitessceConfig,
    CoordinationType as ct,
    Component as cm,
    DataType as dt,
    FileType as ft,
    hconcat,
    vconcat,
)

class TestConfig(unittest.TestCase):

    def test_config_creation(self):
        vc = VitessceConfig()
        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [],
            "coordinationSpace": {},
            "layout": [],
            "initStrategy": "auto"
        })
    
    def test_config_add_dataset(self):
        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
            },
            "layout": [],
            "initStrategy": "auto"
        })
    
    def test_config_add_dataset_add_files(self):
        vc = VitessceConfig()
        my_dataset = (vc.add_dataset(name='My Chained Dataset')
            .add_file(
                url="http://example.com/cells.json",
                data_type=dt.CELLS,
                file_type=ft.CELLS_JSON,
            ).add_file(
                url="http://example.com/cell_sets.json",
                data_type=dt.CELL_SETS,
                file_type=ft.CELL_SETS_JSON,
            )
        )

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Chained Dataset',
                    'files': [
                        {
                            'url': 'http://example.com/cells.json',
                            'type': 'cells',
                            'fileType': 'cells.json'
                        },
                        {
                            'url': 'http://example.com/cell_sets.json',
                            'type': 'cell-sets',
                            'fileType': 'cell-sets.json'
                        }
                    ]
                },
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
            },
            "layout": [],
            "initStrategy": "auto"
        })
    
    def test_config_add_spatial_view(self):
        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')

        my_view = vc.add_view(my_dataset, cm.SPATIAL)

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
            },
            "layout": [
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'h': 1,
                    'w': 1,
                    'x': 0,
                    'y': 0
                }
            ],
            "initStrategy": "auto"
        })
    
    def test_config_add_scatterplot_view_with_mapping(self):
        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')

        my_view = vc.add_view(my_dataset, cm.SCATTERPLOT, mapping="X_umap")

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
                'embeddingType': {
                    'A': 'X_umap'
                }
            },
            "layout": [
                {
                    'component': 'scatterplot',
                    'coordinationScopes': {
                        'dataset': 'A',
                        'embeddingType': 'A'
                    },
                    'h': 1,
                    'w': 1,
                    'x': 0,
                    'y': 0
                }
            ],
            "initStrategy": "auto"
        })
    
    def test_config_add_scatterplot_view_with_embedding_coordinations(self):
        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')

        my_view = vc.add_view(my_dataset, cm.SCATTERPLOT)

        et_scope, ez_scope, ex_scope, ey_scope = vc.add_coordination(ct.EMBEDDING_TYPE, ct.EMBEDDING_ZOOM, ct.EMBEDDING_TARGET_X, ct.EMBEDDING_TARGET_Y)
        my_view.use_coordination(et_scope, ez_scope, ex_scope, ey_scope)

        et_scope.set_value("X_pca")
        ez_scope.set_value(2)
        ex_scope.set_value(10)
        ey_scope.set_value(11)

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
                'embeddingType': {
                    'A': 'X_pca'
                },
                'embeddingZoom': {
                    'A': 2
                },
                'embeddingTargetX': {
                    'A': 10
                },
                'embeddingTargetY': {
                    'A': 11
                },
            },
            "layout": [
                {
                    'component': 'scatterplot',
                    'coordinationScopes': {
                        'dataset': 'A',
                        'embeddingType': 'A',
                        'embeddingZoom': 'A',
                        'embeddingTargetX': 'A',
                        'embeddingTargetY': 'A',
                    },
                    'h': 1,
                    'w': 1,
                    'x': 0,
                    'y': 0
                }
            ],
            "initStrategy": "auto"
        })

    def test_config_add_dataset_add_objects(self):
        vc = VitessceConfig()

        class MockAnnData:
            def __init__(self, name):
                self.name = name
        
        def serve_obj(obj, obj_i, dataset_uid):
            if type(obj) == MockAnnData:
                if obj.name == "Experiment A":
                    return [
                        {
                            "url": "http://localhost:8000/cells",
                            "type": "cells",
                            "fileType": "cells.json"
                        },
                        {
                            "url": "http://localhost:8000/molecules",
                            "type": "molecules",
                            "fileType": "molecules.json"
                        }
                    ]
                elif obj.name == "Experiment B":
                    return [
                        {
                            "url": "http://localhost:8000/cell-sets",
                            "type": "cell-sets",
                            "fileType": "cell-sets.json"
                        }
                    ]
            return None

        my_dataset = (vc.add_dataset(name='My Object Dataset')
            .add_object(
                obj=MockAnnData("Experiment A")
            ).add_object(
                obj=MockAnnData("Experiment B")
            )
        )

        vc_dict = vc.to_dict(on_obj=serve_obj)
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Object Dataset',
                    'files': [
                        {
                            "url": "http://localhost:8000/cells",
                            "type": "cells",
                            "fileType": "cells.json"
                        },
                        {
                            "url": "http://localhost:8000/molecules",
                            "type": "molecules",
                            "fileType": "molecules.json"
                        },
                        {
                            "url": "http://localhost:8000/cell-sets",
                            "type": "cell-sets",
                            "fileType": "cell-sets.json"
                        }
                    ]
                },
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
            },
            "layout": [],
            "initStrategy": "auto"
        })
    
    def test_config_set_layout_single_view(self):
        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')
        my_view = vc.add_view(my_dataset, cm.SPATIAL)
        vc.layout(my_view)

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
            },
            "layout": [
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'x': 0,
                    'y': 0,
                    'h': 12,
                    'w': 12,
                }
            ],
            "initStrategy": "auto"
        })

    def test_config_set_layout_multi_view(self):
        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')
        v1 = vc.add_view(my_dataset, cm.SPATIAL)
        v2 = vc.add_view(my_dataset, cm.SPATIAL)
        v3 = vc.add_view(my_dataset, cm.SPATIAL)

        vc.layout(hconcat(v1, vconcat(v2, v3)))

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
            },
            "layout": [
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'x': 0,
                    'y': 0,
                    'h': 12,
                    'w': 6,
                },
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'x': 6,
                    'y': 0,
                    'h': 6,
                    'w': 6,
                },
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'x': 6,
                    'y': 6,
                    'h': 6,
                    'w': 6,
                }
            ],
            "initStrategy": "auto"
        })
    
    def test_config_set_layout_multi_view_magic(self):
        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')
        v1 = vc.add_view(my_dataset, cm.SPATIAL)
        v2 = vc.add_view(my_dataset, cm.SPATIAL)
        v3 = vc.add_view(my_dataset, cm.SPATIAL)

        vc.layout(v1 | (v2 / v3))

        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "",
            "description": "",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
            },
            "layout": [
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'x': 0,
                    'y': 0,
                    'h': 12,
                    'w': 6,
                },
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'x': 6,
                    'y': 0,
                    'h': 6,
                    'w': 6,
                },
                {
                    'component': 'spatial',
                    'coordinationScopes': {
                        'dataset': 'A',
                    },
                    'x': 6,
                    'y': 6,
                    'h': 6,
                    'w': 6,
                }
            ],
            "initStrategy": "auto"
        })

    def test_load_config(self):
        vc = VitessceConfig.from_dict({
            "version": "1.0.0",
            "name": "Test name",
            "description": "Test description",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My First Dataset',
                    'files': [
                        {
                            'url': 'http://cells.json',
                            'type': 'cells',
                            'fileType': 'cells.json'
                        }
                    ]
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
                'spatialZoom': {
                    'ABC': 11
                },
            },
            "layout": [
                {
                    "component": "spatial",
                    "props": {
                        "cellRadius": 50
                    },
                    "coordinationScopes": {
                        "spatialZoom": 'ABC'
                    },
                    "x": 5,
                    "y": 0,
                    "w": 4,
                    "h": 4
                },
            ],
            "initStrategy": "auto"
        })

        my_second_dataset = vc.add_dataset(name='My Second Dataset')
        
        vc_dict = vc.to_dict()
        vc_json = json.dumps(vc_dict)

        self.assertEqual(vc_dict, {
            "version": "1.0.0",
            "name": "Test name",
            "description": "Test description",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My First Dataset',
                    'files': [
                        {
                            'url': 'http://cells.json',
                            'type': 'cells',
                            'fileType': 'cells.json'
                        }
                    ]
                },
                {
                    'uid': 'B',
                    'name': 'My Second Dataset',
                    'files': []
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A',
                    'B': 'B'
                },
                'spatialZoom': {
                    'ABC': 11,
                },
            },
            "layout": [
                {
                    "component": "spatial",
                    "props": {
                        "cellRadius": 50
                    },
                    "coordinationScopes": {
                        "spatialZoom": 'ABC'
                    },
                    "x": 5,
                    "y": 0,
                    "w": 4,
                    "h": 4
                },
            ],
            "initStrategy": "auto"
        })




from ._version import version_info, __version__

from .widget import VitessceWidget
from .config import VitessceConfig, hconcat, vconcat
from .constants import CoordinationType, Component, DataType, FileType
from .wrappers import (
    AbstractWrapper,
    OmeTiffWrapper,
    OmeZarrWrapper,
    AnnDataWrapper,
    LoomWrapper,
    SnapWrapper,
)
from .entities import (
    CellSets,
    Cells,
    Molecules,
)

def _jupyter_nbextension_paths():
    """Called by Jupyter Notebook Server to detect if it is a valid nbextension and
    to install the widget

    Returns
    =======
    section: The section of the Jupyter Notebook Server to change.
        Must be 'notebook' for widget extensions
    src: Source directory name to copy files from. Webpack outputs generated files
        into this directory and Jupyter Notebook copies from this directory during
        widget installation
    dest: Destination directory name to install widget files to. Jupyter Notebook copies
        from `src` directory into <jupyter path>/nbextensions/<dest> directory
        during widget installation
    require: Path to importable AMD Javascript module inside the
        <jupyter path>/nbextensions/<dest> directory
    """
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'vitessce-jupyter',
        'require': 'vitessce-jupyter/extension'
    }]



import unittest

from vitessce import (
    CellSets,
    Cells,
)

class TestEntities(unittest.TestCase):

    def test_cells(self):

      cell_ids = ['cell_1', 'cell_2', 'cell_3']

      cells = Cells(cell_ids=cell_ids)
      self.assertEqual(list(cells.json.keys()), cell_ids)

      cells.add_mapping('umap', [[1, 1], [2, 2], [3, 3]])

      cells.add_mapping('pca', [[1, 1], [2, 2], [3, 3]])

      cells.add_centroids([[1, 1], [2, 2], [3, 3]])

      cells.add_polygon_outline([[[1, 1], [1, 1], [1, 1]], [[2, 2], [2, 2], [2, 2]], [[3, 3], [3, 3], [3, 3]]])

      self.assertEqual(
        cells.json,
        {
          'cell_1': {
            'mappings': { 'umap': [1, 1], 'pca': [1, 1] },
            'xy': [1, 1],
            'poly': [[1, 1], [1, 1], [1, 1]]
          },
          'cell_2': {
            'mappings': { 'umap': [2, 2], 'pca': [2, 2] },
            'xy': [2, 2],
            'poly': [[2, 2], [2, 2], [2, 2]]
          },
          'cell_3': {
            'mappings': { 'umap': [3, 3], 'pca': [3, 3] },
            'xy': [3, 3],
            'poly': [[3, 3], [3, 3], [3, 3]]
          }
        }
      )
    
    def test_cells_bad_polygon_outline_type(self):

      cell_ids = ['cell_1', 'cell_2', 'cell_3']
      cells = Cells(cell_ids=cell_ids)
      with self.assertRaises(Exception) as context:
        # The extra 3 should be problematic since polygons are two dimensional.
        cells.add_polygon_outline([[[1, 1, 3], [1, 1], [1, 1]], [[2, 2], [2, 2], [2, 2]], [[3, 3], [3, 3], [3, 3]]])
      self.assertEqual('Polygon outline for cell_1 should be a list of two element lists i.e xy coordinates', str(context.exception))
    
    def test_cells_bad_mappings_length(self):

      cell_ids = ['cell_1', 'cell_2', 'cell_3']
      cells = Cells(cell_ids=cell_ids)
      with self.assertRaises(Exception) as context:
        # There are 3 cells in this object so only passing in two scatterplot cooridnates is problematic.
        cells.add_mapping('umap', [[1, 1], [2, 2]])
      self.assertEqual('Coordinates length does not match Cell IDs Length', str(context.exception))

    def test_cell_sets(self):

      cell_sets = CellSets()
      cell_sets.add_level_zero_node('Clusters')
      
      cell_sets.add_node('Cluster 1', ['Clusters'])
      cell_sets.add_node('Cluster 2', ['Clusters'])
      cell_sets.add_node('Subcluster 1', ['Clusters', 'Cluster 1'], ['cell_1', 'cell_2'])
      cell_sets.add_node('Subcluster 2', ['Clusters', 'Cluster 1'], ['cell_3', 'cell_4'])

      self.assertEqual(
        cell_sets.json,
        {
          "datatype": "cell",
          "version": "0.1.2",
          "tree": [{
              "name": 'Clusters',
              "children": [
                {
                  "name": 'Cluster 1',
                  "children": [
                    {
                      "name": 'Subcluster 1',
                      "set": ['cell_1', 'cell_2']
                    },
                    {
                      "name": 'Subcluster 2',
                      "set": ['cell_3', 'cell_4']
                    }
                  ]
                },
                {
                  "name": 'Cluster 2',
                }
              ]
          }]
        }
      )

    def test_cell_sets_node_not_found(self):

      cell_sets = CellSets()
      with self.assertRaises(Exception) as context:
        cell_sets.add_node('Cluster 1', ['Clusters Not Found'])

      self.assertEqual("No node with path ['Clusters Not Found'] found to add Cluster 1 to", str(context.exception))


import os
from os.path import join
import tempfile
import math

import numpy as np
import pandas as pd
import zarr
from numcodecs import Zlib
from scipy.sparse import csr_matrix
from scipy.sparse import coo_matrix

from starlette.responses import JSONResponse, UJSONResponse
from starlette.routing import Route, Mount
from starlette.staticfiles import StaticFiles

from .constants import DataType as dt, FileType as ft
from .entities import Cells, CellSets, GenomicProfiles

class AbstractWrapper:
    """
    An abstract class that can be extended when
    implementing custom dataset object wrapper classes. 
    """

    def __init__(self, **kwargs):
        """
        Abstract constructor to be inherited by dataset wrapper classes.

        :param str base_url: An optional base URL to use in dataset file definitions.
        """
        self._base_url = kwargs['base_url'] if 'base_url' in kwargs else None

    def get_cells(self, port, dataset_uid, obj_i):
        """
        Get the file definitions and server routes
        corresponding to the :class:`~vitessce.constants.DataType.CELLS` data type.
        Used internally by :class:`~vitessce.widget.VitessceWidget`.

        :param int port: The web server port, meant to be used in the localhost URLs in the file definitions.
        :param str dataset_uid: The unique identifier for the dataset parent of this data object.
        :param int obj_i: The index of this data object child within its dataset parent.

        :returns: The file definitions and server routes.
        :rtype: tuple[list[dict], list[starlette.routing.Route]]
        """
        raise NotImplementedError()

    def get_cell_sets(self, port, dataset_uid, obj_i):
        """
        Get the file definitions and server routes
        corresponding to the :class:`~vitessce.constants.DataType.CELL_SETS` data type.
        Used internally by :class:`~vitessce.widget.VitessceWidget`.

        :param int port: The web server port, meant to be used in the localhost URLs in the file definitions.
        :param str dataset_uid: The unique identifier for the dataset parent of this data object.
        :param int obj_i: The index of this data object child within its dataset parent.

        :returns: The file definitions and server routes.
        :rtype: tuple[list[dict], list[starlette.routing.Route]]
        """
        raise NotImplementedError()

    def get_raster(self, port, dataset_uid, obj_i):
        """
        Get the file definitions and server routes
        corresponding to the :class:`~vitessce.constants.DataType.RASTER` data type.
        Used internally by :class:`~vitessce.widget.VitessceWidget`.

        :param int port: The web server port, meant to be used in the localhost URLs in the file definitions.
        :param str dataset_uid: The unique identifier for the dataset parent of this data object.
        :param int obj_i: The index of this data object child within its dataset parent.

        :returns: The file definitions and server routes.
        :rtype: tuple[list[dict], list[starlette.routing.Route]]
        """
        raise NotImplementedError()

    def get_molecules(self, port, dataset_uid, obj_i):
        """
        Get the file definitions and server routes
        corresponding to the :class:`~vitessce.constants.DataType.MOLECULES` data type.

        :param int port: The web server port, meant to be used in the localhost URLs in the file definitions.
        :param str dataset_uid: The unique identifier for the dataset parent of this data object.
        :param int obj_i: The index of this data object child within its dataset parent.

        :returns: The file definitions and server routes.
        :rtype: tuple[list[dict], list[starlette.routing.Route]]
        """
        raise NotImplementedError()

    def get_neighborhoods(self, port, dataset_uid, obj_i):
        """
        Get the file definitions and server routes
        corresponding to the :class:`~vitessce.constants.DataType.NEIGHBORHOODS` data type.
        Used internally by :class:`~vitessce.widget.VitessceWidget`.

        :param int port: The web server port, meant to be used in the localhost URLs in the file definitions.
        :param str dataset_uid: The unique identifier for the dataset parent of this data object.
        :param int obj_i: The index of this data object child within its dataset parent.

        :returns: The file definitions and server routes.
        :rtype: tuple[list[dict], list[starlette.routing.Route]]
        """
        raise NotImplementedError()

    def get_expression_matrix(self, port, dataset_uid, obj_i):
        """
        Get the file definitions and server routes
        corresponding to the :class:`~vitessce.constants.DataType.EXPRESSION_MATRIX` data type.
        Used internally by :class:`~vitessce.widget.VitessceWidget`.

        :param int port: The web server port, meant to be used in the localhost URLs in the file definitions.
        :param str dataset_uid: The unique identifier for the dataset parent of this data object.
        :param int obj_i: The index of this data object child within its dataset parent.

        :returns: The file definitions and server routes.
        :rtype: tuple[list[dict], list[starlette.routing.Route]]
        """
        raise NotImplementedError()

    def get_genomic_profiles(self, port, dataset_uid, obj_i):
        """
        Get the file definitions and server routes
        corresponding to the :class:`~vitessce.constants.DataType.GENOMIC_PROFILES` data type.
        Used internally by :class:`~vitessce.widget.VitessceWidget`.

        :param int port: The web server port, meant to be used in the localhost URLs in the file definitions.
        :param str dataset_uid: The unique identifier for the dataset parent of this data object.
        :param int obj_i: The index of this data object child within its dataset parent.

        :returns: The file definitions and server routes.
        :rtype: tuple[list[dict], list[starlette.routing.Route]]
        """
        raise NotImplementedError()

    def _create_response_json(self, data_json):
        """
        Helper function that can be used for creating JSON responses.

        :param dict data_json: The data to return as JSON in the response body.
        :returns: The response handler function.
        :rtype: function
        """
        async def response_func(req):
            return UJSONResponse(data_json)
        return response_func

    def _get_data(self, data_type, port, dataset_uid, obj_i):
        if data_type == dt.CELLS:
            return self.get_cells(port, dataset_uid, obj_i)
        elif data_type == dt.CELL_SETS:
            return self.get_cell_sets(port, dataset_uid, obj_i)
        elif data_type == dt.RASTER:
            return self.get_raster(port, dataset_uid, obj_i)
        elif data_type == dt.MOLECULES:
            return self.get_molecules(port, dataset_uid, obj_i)
        elif data_type == dt.NEIGHBORHOODS:
            return self.get_neighborhoods(port, dataset_uid, obj_i)
        elif data_type == dt.EXPRESSION_MATRIX:
            return self.get_expression_matrix(port, dataset_uid, obj_i)
        elif data_type == dt.GENOMIC_PROFILES:
            return self.get_genomic_profiles(port, dataset_uid, obj_i)

    def _get_url(self, port, dataset_uid, obj_i, suffix):
        # A base URL is defined for this so this is used outside of Jupyter notebook.
        if self._base_url is not None:
            return f"{self._base_url}/{dataset_uid}/{obj_i}/{suffix}"
        return f"http://localhost:{port}/{dataset_uid}/{obj_i}/{suffix}"

    def _get_route(self, dataset_uid, obj_i, suffix):
        return f"/{dataset_uid}/{obj_i}/{suffix}"


class OmeTiffWrapper(AbstractWrapper):

    def __init__(self, img_path, offsets_path=None, name="", **kwargs):
        super().__init__(**kwargs)
        self.img_path = img_path
        self.offsets_path = offsets_path
        self.name = name

    def create_raster_json(self, img_url, offsets_url):
        raster_json = {
            "schemaVersion": "0.0.2",
            "images": [
                {
                    "name": self.name,
                    "type": "ome-tiff",
                    "url": img_url,
                    "metadata": {
                        **({
                            "omeTiffOffsetsUrl": offsets_url,
                        } if offsets_url is not None else {})
                    }
                }
            ],
        }
        return raster_json

    def _get_offsets_dir(self):
        return os.path.dirname(self.offsets_path)
    
    def _get_offsets_filename(self):
        return os.path.basename(self.offsets_path)

    def get_raster(self, port, dataset_uid, obj_i):
        img_dir_path, img_url = self.img_path, self._get_url(port, dataset_uid, obj_i, "raster_img")
        offsets_dir_path, offsets_url = (None, None) if self.offsets_path is None else (self._get_offsets_dir(), self._get_url(port, dataset_uid, obj_i, join("raster_offsets", self._get_offsets_filename())))

        raster_json = self.create_raster_json(img_url, offsets_url)

        obj_routes = [
            Mount(self._get_route(dataset_uid, obj_i, "raster_img"),
                  app=StaticFiles(directory=img_dir_path, html=False, check_dir=False)),
            Route(self._get_route(dataset_uid, obj_i, "raster"),
                  self._create_response_json(raster_json))
        ]
        if self.offsets_path is not None:
            obj_routes.append(
                Mount(self._get_route(dataset_uid, obj_i, "raster_offsets"),
                      app=StaticFiles(directory=offsets_dir_path, html=False, check_dir=False))
            )

        obj_file_defs = [
            {
                "type": dt.RASTER.value,
                "fileType": ft.RASTER_JSON.value,
                "url": self._get_url(port, dataset_uid, obj_i, "raster")
            }
        ]

        return obj_file_defs, obj_routes


class OmeZarrWrapper(AbstractWrapper):

    def __init__(self, z, name="", **kwargs):
        super().__init__(**kwargs)
        self.z = z
        self.name = name

    def create_raster_json(self, img_url):
        raster_json = {
            "schemaVersion": "0.0.2",
            "images": [
                {
                    "name": self.name,
                    "type": "zarr",
                    "url": img_url,
                    "metadata": {
                        "dimensions": [
                            {
                                "field": "channel",
                                "type": "nominal",
                                "values": [
                                    "DAPI - Hoechst (nuclei)",
                                    "FITC - Laminin (basement membrane)",
                                    "Cy3 - Synaptopodin (glomerular)",
                                    "Cy5 - THP (thick limb)"
                                ]
                            },
                            {
                                "field": "y",
                                "type": "quantitative",
                                "values": None
                            },
                            {
                                "field": "x",
                                "type": "quantitative",
                                "values": None
                            }
                        ],
                        "isPyramid": True,
                        "transform": {
                            "scale": 1,
                            "translate": {
                                "x": 0,
                                "y": 0,
                            }
                        }
                    }
                }
            ],
        }
        return raster_json

    def get_raster(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []

        if type(self.z) == zarr.hierarchy.Group:
            img_dir_path = self.z.store.path

            raster_json = self.create_raster_json(
                self._get_url(port, dataset_uid, obj_i, "raster_img"),
            )

            obj_routes = [
                Mount(self._get_route(dataset_uid, obj_i, "raster_img"),
                        app=StaticFiles(directory=img_dir_path, html=False)),
                Route(self._get_route(dataset_uid, obj_i, "raster"),
                        self._create_response_json(raster_json))
            ]
            obj_file_defs = [
                {
                    "type": dt.RASTER.value,
                    "fileType": ft.RASTER_JSON.value,
                    "url": self._get_url(port, dataset_uid, obj_i, "raster")
                }
            ]

        return obj_file_defs, obj_routes


class AnnDataWrapper(AbstractWrapper):
    def __init__(self, adata, use_highly_variable_genes=True, cell_set_obs_cols=None, **kwargs):
        """
        Wrap an AnnData object by creating an instance of the ``AnnDataWrapper`` class.

        :param adata: An AnnData object containing single-cell experiment data.
        :type adata: anndata.AnnData
        :param bool use_highly_variable_genes: When creating outputs with genes, should only the genes marked as highly variable be used?
        :param list[str] cell_set_obs_cols: A list of column names of the ``adata.obs`` dataframe that should be used for creating cell sets.
        :param \*\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
        """
        super().__init__(**kwargs)
        self.adata = adata
        self.tempdir = tempfile.mkdtemp()
        self.use_highly_variable_genes = use_highly_variable_genes
        self.cell_set_obs_cols = cell_set_obs_cols

    def create_cells_json(self):
        adata = self.adata
        available_embeddings = list(adata.obsm.keys())

        cell_ids = adata.obs.index.tolist()
        cells = Cells(cell_ids=cell_ids)
        for e in available_embeddings:
            mapping = adata.obsm[e][:, 0:2].tolist()
            cells.add_mapping(e, mapping)
        return cells.json

    def create_cell_sets_json(self):
        adata = self.adata

        cell_set_obs_cols = self.cell_set_obs_cols

        cell_sets = CellSets()

        if cell_set_obs_cols is not None and len(cell_set_obs_cols) > 0:
            # Each `cell_set_obs_col` is a column name in the `adata.obs` dataframe,
            # which we want to turn into a hierarchy of cell sets.
            for cell_set_obs_col in cell_set_obs_cols:
                cell_sets.add_level_zero_node(cell_set_obs_col)

                cell_ids = adata.obs.index.tolist()
                cluster_ids = adata.obs[cell_set_obs_col].unique().tolist()
                cell_cluster_ids = adata.obs[cell_set_obs_col].values.tolist()

                cell_cluster_tuples = list(zip(cell_ids, cell_cluster_ids))

                for cluster_id in sorted(cluster_ids):
                    cell_set = [
                        str(cell_id)
                        for cell_id, cell_cluster_id in cell_cluster_tuples
                        if cell_cluster_id == cluster_id
                    ]
                    cell_sets.add_node(str(cluster_id), [cell_set_obs_col], cell_set)
            return cell_sets.json
        return None
    
    def create_exp_matrix_zarr(self, zarr_filepath):
        adata = self.adata
        gexp_arr = adata.X

        cell_list = adata.obs.index.values.tolist()
        gene_list = adata.var.index.values.tolist()

        if type(gexp_arr) == csr_matrix:
            # Convert from SciPy sparse format to NumPy dense format
            gexp_arr = gexp_arr.toarray()
        
        if self.use_highly_variable_genes and 'highly_variable' in adata.var.columns.values.tolist():
            # Restrict the gene expression matrix to only the genes marked as highly variable
            gene_list = adata.var.index[adata.var['highly_variable']].values.tolist()
            gexp_arr = gexp_arr[:,adata.var['highly_variable'].values]

        
        # Re-scale the gene expression values between 0 and 255
        gexp_arr_min = gexp_arr.min()
        gexp_arr_max = gexp_arr.max()
        gexp_arr_range = gexp_arr_max - gexp_arr_min
        gexp_arr_ratio = 255 / gexp_arr_range

        gexp_norm_arr = (gexp_arr - gexp_arr_min) * gexp_arr_ratio
    
        z = zarr.open(
            zarr_filepath,
            mode='w',
            shape=gexp_norm_arr.shape,
            dtype='uint8',
            compressor=Zlib(level=1)
        )

        z[:] = gexp_norm_arr
        # observations: cells (rows)
        z.attrs["rows"] = cell_list
        # variables: genes (columns)
        z.attrs["cols"] = gene_list
        
        return

    def get_cells(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []

        cells_json = self.create_cells_json()

        obj_routes = [
            Route(self._get_route(dataset_uid, obj_i, "cells"),
                    self._create_response_json(cells_json)),
        ]
        obj_file_defs = [
            {
                "type": dt.CELLS.value,
                "fileType": ft.CELLS_JSON.value,
                "url": self._get_url(port, dataset_uid, obj_i, "cells")
            }
        ]

        return obj_file_defs, obj_routes

    def get_cell_sets(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []

            
        cell_sets_json = self.create_cell_sets_json()

        if cell_sets_json is not None:
            obj_routes = [
                Route(self._get_route(dataset_uid, obj_i, "cell-sets"),
                        self._create_response_json(cell_sets_json)),
            ]
            obj_file_defs = [
                {
                    "type": dt.CELL_SETS.value,
                    "fileType": ft.CELL_SETS_JSON.value,
                    "url": self._get_url(port, dataset_uid, obj_i, "cell-sets")
                }
            ]

        return obj_file_defs, obj_routes
    
    def get_expression_matrix(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []

        zarr_tempdir = self.tempdir
        zarr_filepath = join(zarr_tempdir, 'matrix.zarr')

        self.create_exp_matrix_zarr(zarr_filepath)

        if zarr_tempdir is not None:
            obj_routes = [
                Mount(self._get_route(dataset_uid, obj_i, "expression"),
                    app=StaticFiles(directory=os.path.dirname(zarr_filepath), html=False, check_dir=False)),
            ]

            obj_file_defs = [
                {
                    "type": dt.EXPRESSION_MATRIX.value,
                    "fileType": ft.EXPRESSION_MATRIX_ZARR.value,
                    "url": self._get_url(port, dataset_uid, obj_i, "expression/matrix.zarr")
                }
            ]

        return obj_file_defs, obj_routes
        


class LoomWrapper(AbstractWrapper):

    def __init__(self, loom, **kwargs):
        super().__init__(**kwargs)
        self.loom = loom

    def get_cells(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []

        # TODO: append routes
        # TODO: add file definitions
 
        return obj_file_defs, obj_routes

class SnapWrapper(AbstractWrapper):

    # The Snap file is difficult to work with.
    # For now we can use the processed cell-by-bin MTX file
    # However, the HuBMAP pipeline currently computes this with resolution 5000
    # TODO: Make a PR to sc-atac-seq-pipeline to output this at a higher resolution (e.g. 200)
    # https://github.com/hubmapconsortium/sc-atac-seq-pipeline/blob/develop/bin/snapAnalysis.R#L93

    def __init__(self, in_mtx, in_barcodes_df, in_bins_df, in_clusters_df, starting_resolution=5000, **kwargs):
        super().__init__(**kwargs)
        self.in_mtx = in_mtx # scipy.sparse.coo.coo_matrix (filtered_cell_by_bin.mtx)
        self.in_barcodes_df = in_barcodes_df # pandas dataframe (barcodes.txt)
        self.in_bins_df = in_bins_df # pandas dataframe (bins.txt)
        self.in_clusters_df = in_clusters_df # pandas dataframe (umap_coords_clusters.csv)

        self.tempdir = tempfile.mkdtemp()

        self.starting_resolution = starting_resolution

        # Convert to dense matrix if sparse.
        if type(in_mtx) == coo_matrix:
            self.in_mtx = in_mtx.toarray()


    def create_genomic_multivec_zarr(self, zarr_filepath):
        in_mtx = self.in_mtx
        in_clusters_df = self.in_clusters_df
        in_barcodes_df = self.in_barcodes_df
        in_bins_df = self.in_bins_df

        starting_resolution = self.starting_resolution

        # The bin datafram consists of one column like chrName:binStart-binEnd
        def convert_bin_name_to_chr_name(bin_name):
            try:
                return bin_name[:bin_name.index(':')]
            except ValueError:
                return np.nan
        def convert_bin_name_to_chr_start(bin_name):
            try:
                return int(bin_name[bin_name.index(':')+1:bin_name.index('-')])
            except ValueError:
                return np.nan
        def convert_bin_name_to_chr_end(bin_name):
            try:
                return int(bin_name[bin_name.index('-')+1:])
            except ValueError:
                return np.nan
        
        # The genome assembly is GRCh38 but the chromosome names in the bin names do not start with the "chr" prefix.
        # This is incompatible with the chromosome names from `negspy`, so we need to append the prefix.
        in_bins_df[0] = in_bins_df[0].apply(lambda x: "chr" + x)
        
        in_bins_df["chr_name"] = in_bins_df[0].apply(convert_bin_name_to_chr_name)
        in_bins_df["chr_start"] = in_bins_df[0].apply(convert_bin_name_to_chr_start)
        in_bins_df["chr_end"] = in_bins_df[0].apply(convert_bin_name_to_chr_end)

        # Drop any rows that had incorrect bin strings (missing a chromosome name, bin start, or bin end value).
        in_bins_df = in_bins_df.dropna(subset=["chr_name", "chr_start", "chr_end"]).copy()

        # Ensure that the columns have the expect types.
        in_bins_df["chr_name"] = in_bins_df["chr_name"].astype(str)
        in_bins_df["chr_start"] = in_bins_df["chr_start"].astype(int)
        in_bins_df["chr_end"] = in_bins_df["chr_end"].astype(int)

        # Create the Zarr store for the outputs.
        out_f = zarr.open(zarr_filepath, mode='w')

        # Get a list of clusters.
        in_clusters_df["cluster"] = in_clusters_df["cluster"].astype(str)
        cluster_ids = in_clusters_df["cluster"].unique().tolist()
        cluster_ids.sort(key=int)

        cluster_paths = [ [ "Clusters", cluster_id ] for cluster_id in cluster_ids ]
        
        # "SnapTools performs quantification using a specified aligner, and HuBMAP has standardized on BWA with the GRCh38 reference genome"
        # Reference: https://github.com/hubmapconsortium/sc-atac-seq-pipeline/blob/bb023f95ca3330128bfef41cc719ffcb2ee6a190/README.md
        genomic_profiles = GenomicProfiles(out_f, profile_paths=cluster_paths, assembly='hg38', starting_resolution=starting_resolution)
        chrom_name_to_length = genomic_profiles.chrom_name_to_length

        # Create each chromosome dataset.
        for chr_name, chr_len in chrom_name_to_length.items():
            # The bins dataframe frustratingly does not contain every bin.
            # We need to figure out which bins are missing.

            # We want to check for missing bins in each chromosome separately,
            # otherwise too much memory is used during the join step.
            chr_bins_in_df = in_bins_df.loc[in_bins_df["chr_name"] == chr_name]
            if chr_bins_in_df.shape[0] == 0:
                # No processing or output is necessary if there is no data for this chromosome.
                # Continue on through all resolutions of this chromosome to the next chromosome.
                continue
            
            # Determine the indices of the matrix at which the bins for this chromosome start and end.
            chr_bin_i_start = int(chr_bins_in_df.head(1).iloc[0].name)
            chr_bin_i_end = int(chr_bins_in_df.tail(1).iloc[0].name) + 1
            
            # Extract the part of the matrix corresponding to the current chromosome.
            chr_mtx = in_mtx[:,chr_bin_i_start:chr_bin_i_end]

            # Create a list of the "ground truth" bins (all bins from position 0 to the end of the chromosome).
            # We will join the input bins onto this dataframe to determine which bins are missing.
            chr_bins_gt_df = pd.DataFrame()
            chr_bins_gt_df["chr_start"] = np.arange(0, math.ceil(chr_len/starting_resolution)) * starting_resolution
            chr_bins_gt_df["chr_end"] = chr_bins_gt_df["chr_start"] + starting_resolution
            chr_bins_gt_df["chr_start"] = chr_bins_gt_df["chr_start"] + 1
            chr_bins_gt_df["chr_start"] = chr_bins_gt_df["chr_start"].astype(int)
            chr_bins_gt_df["chr_end"] = chr_bins_gt_df["chr_end"].astype(int)
            chr_bins_gt_df["chr_name"] = chr_name
            chr_bins_gt_df[0] = chr_bins_gt_df.apply(lambda r: f"{r['chr_name']}:{r['chr_start']}-{r['chr_end']}", axis='columns')
            
            # We will add a new column "i", which should match the _old_ index, so that we will be able join with the data matrix on the original indices.
            # For the new rows, we will add values for the "i" column that are greater than any of the original indices,
            # to prevent any joining with the incoming data matrix onto these bins for which the data is missing.
            chr_bins_in_df = chr_bins_in_df.reset_index(drop=True)
            chr_bins_in_df["i"] = chr_bins_in_df.index.values
            chr_bins_gt_df["i"] = chr_bins_gt_df.index.values + (in_mtx.shape[1] + 1)
            
            # Set the full bin string column as the index of both data frames.
            chr_bins_gt_df = chr_bins_gt_df.set_index(0)
            chr_bins_in_df = chr_bins_in_df.set_index(0)
            
            # Join the input bin subset dataframe right onto the full bin ground truth dataframe.
            chr_bins_in_join_df = chr_bins_in_df.join(chr_bins_gt_df, how='right', lsuffix="", rsuffix="_gt")
            # The bins which were not present in the input will have NaN values in the "i" column.
            # For these rows, we replace the NaN values with the much higher "i_gt" values which will not match to any index of the data matrix.
            chr_bins_in_join_df["i"] = chr_bins_in_join_df.apply(lambda r: r['i'] if pd.notna(r['i']) else r['i_gt'], axis='columns').astype(int)

            # Clean up the joined data frame by removing unnecessary columns.
            chr_bins_in_join_df = chr_bins_in_join_df.drop(columns=['chr_name', 'chr_start', 'chr_end', 'i_gt'])
            chr_bins_in_join_df = chr_bins_in_join_df.rename(columns={'chr_name_gt': 'chr_name', 'chr_start_gt': 'chr_start', 'chr_end_gt': 'chr_end'})
            
            # Create a dataframe from the data matrix, so that we can join to the joined bins dataframe.
            chr_mtx_df = pd.DataFrame(data=chr_mtx.T)
            
            chr_bins_i_df = chr_bins_in_join_df.drop(columns=['chr_name', 'chr_start', 'chr_end'])

            # Join the data matrix dataframe and the bins dataframe.
            # Bins that are missing from the data matrix will have "i" values higher than any of the data matrix dataframe row indices,
            # and therefore the data values for these bins in the resulting joined dataframe will all be NaN.
            chr_mtx_join_df = chr_bins_i_df.join(chr_mtx_df, how='left', on='i')
            # We fill in these NaN values with 0.
            chr_mtx_join_df = chr_mtx_join_df.fillna(value=0.0)
            
            # Drop the "i" column, since it is not necessary now that we have done the join.
            chr_mtx_join_df = chr_mtx_join_df.drop(columns=['i'])
            # Obtain the new full data matrix, which contains values for all bins of the chromosome.
            chr_mtx = chr_mtx_join_df.values.T

            # Fill in the Zarr store with data for each cluster.
            for cluster_index, cluster_id in enumerate(cluster_ids):
                # Get the list of cells in the current cluster.
                cluster_df = in_clusters_df.loc[in_clusters_df["cluster"] == cluster_id]
                cluster_cell_ids = cluster_df.index.values.tolist()
                cluster_num_cells = len(cluster_cell_ids)
                cluster_cells_tf = (in_barcodes_df[0].isin(cluster_cell_ids)).values

                # Get the rows of the data matrix corresponding to the cells in this cluster.
                cluster_cell_by_bin_mtx = chr_mtx[cluster_cells_tf,:]
                # Take the sum of this cluster along the cells axis.
                cluster_profile = cluster_cell_by_bin_mtx.sum(axis=0)

                genomic_profiles.add_profile(cluster_profile, chr_name, cluster_index)
        
        return

    def get_genomic_profiles(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []
        
        zarr_tempdir = self.tempdir
        zarr_filepath = join(zarr_tempdir, 'profiles.zarr')

        print("Please wait, the following conversion is slow")
        self.create_genomic_multivec_zarr(zarr_filepath)

        if zarr_tempdir is not None:
            obj_routes = [
                Mount(self._get_route(dataset_uid, obj_i, "genomic"),
                    app=StaticFiles(directory=os.path.dirname(zarr_filepath), html=False, check_dir=False)),
            ]

            obj_file_defs = [
                {
                    "type": dt.GENOMIC_PROFILES.value,
                    "fileType": ft.GENOMIC_PROFILES_ZARR.value,
                    "url": self._get_url(port, dataset_uid, obj_i, "genomic/profiles.zarr")
                }
            ]

        return obj_file_defs, obj_routes
    

    def create_cell_sets_json(self):
        in_clusters_df = self.in_clusters_df
        cell_sets = CellSets()
        cell_sets.add_level_zero_node('Clusters')

        cell_ids = in_clusters_df.index.values.tolist()
        in_clusters_df['cluster'] = in_clusters_df['cluster'].astype(str)
        cluster_ids = in_clusters_df['cluster'].unique().tolist()
        cluster_ids.sort(key=int)
        cell_cluster_ids = in_clusters_df['cluster'].values.tolist()

        cell_cluster_tuples = list(zip(cell_ids, cell_cluster_ids))

        for cluster_id in cluster_ids:
            cell_set = [
                str(cell_id)
                for cell_id, cell_cluster_id in cell_cluster_tuples
                if cell_cluster_id == cluster_id
            ]
            cell_sets.add_node(str(cluster_id), ['Clusters'], cell_set)

        return cell_sets.json
    
    def get_cell_sets(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []

        cell_sets_json = self.create_cell_sets_json()

        obj_routes = [
            Route(self._get_route(dataset_uid, obj_i, "cell-sets"),
                    self._create_response_json(cell_sets_json)),
        ]
        obj_file_defs = [
            {
                "type": dt.CELL_SETS.value,
                "fileType": ft.CELL_SETS_JSON.value,
                "url": self._get_url(port, dataset_uid, obj_i, "cell-sets")
            }
        ]

        return obj_file_defs, obj_routes
    
    def create_cells_json(self):
        in_clusters_df = self.in_clusters_df

        cell_ids = in_clusters_df.index.tolist()
        cells = Cells(cell_ids=cell_ids)
        mapping = in_clusters_df[["umap.1", "umap.2"]].values.tolist()
        cells.add_mapping("UMAP", mapping)
        return cells.json
    
    def get_cells(self, port, dataset_uid, obj_i):
        obj_routes = []
        obj_file_defs = []

        cells_json = self.create_cells_json()

        obj_routes = [
            Route(self._get_route(dataset_uid, obj_i, "cells"),
                    self._create_response_json(cells_json)),
        ]
        obj_file_defs = [
            {
                "type": dt.CELLS.value,
                "fileType": ft.CELLS_JSON.value,
                "url": self._get_url(port, dataset_uid, obj_i, "cells")
            }
        ]

        return obj_file_defs, obj_routes



from enum import Enum

# Reference: https://stackoverflow.com/a/50473952
class DocEnum(Enum):
    def __new__(cls, value, doc=None):
        self = object.__new__(cls)
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self

class CoordinationType(DocEnum):
    """
    An enum type representing a coordination type in the Vitessce coordination model.
    The term coordination type refers to a parameter to be coordinated, and its programming-language-like type.
    For example, the ``SPATIAL_ZOOM`` coordination type represents a coordination of the zoom level of a spatial view, which can take a float value.
    """
    DATASET = "dataset", "The identifier for the dataset associated with a view."
    EMBEDDING_TYPE = "embeddingType", "The type of embedding used for a scatterplot view, such as PCA or t-SNE."
    EMBEDDING_ZOOM = "embeddingZoom", "The zoom level of an embedding scatterplot view."
    EMBEDDING_ROTATION = "embeddingRotation", "The rotation of an embedding scatterplot view."
    EMBEDDING_TARGET_X = "embeddingTargetX", "The x-coordinate of the center of an embedding scatterplot view."
    EMBEDDING_TARGET_Y = "embeddingTargetY", "The y-coordinate of the center of an embedding scatterplot view."
    EMBEDDING_TARGET_Z = "embeddingTargetZ", "The z-coordinate of the center of an embedding scatterplot view."
    SPATIAL_ZOOM = "spatialZoom", "The zoom level of a spatial view."
    SPATIAL_ROTATION = "spatialRotation", "The rotation of a spatial view."
    SPATIAL_TARGET_X = "spatialTargetX", "The x-coordinate of the center of a spatial view."
    SPATIAL_TARGET_Y = "spatialTargetY", "The y-coordinate of the center of a spatial view."
    SPATIAL_TARGET_Z = "spatialTargetZ", "The z-coordinate of the center of a spatial view."
    HEATMAP_ZOOM_X = "heatmapZoomX", "The x-axis zoom level of a heatmap view."
    HEATMAP_ZOOM_Y = "heatmapZoomY", "The y-axis zoom level of a heatmap view."
    HEATMAP_TARGET_X = "heatmapTargetX", "The x-coordinate of the center of a heatmap view."
    HEATMAP_TARGET_Y = "heatmapTargetY", "The y-coordinate of the center of a heatmap view."
    CELL_FILTER = "cellFilter", "A subset of cells to include after filtering."
    CELL_HIGHLIGHT = "cellHighlight", "A subset of cells to highlight."
    CELL_SET_SELECTION = "cellSetSelection", "A subset of cell sets to select."
    CELL_SET_HIGHLIGHT = "cellSetHighlight", "A subset of cell sets to highlight."
    CELL_SET_COLOR = "cellSetColor", "A mapping from cell sets to colors."
    GENE_FILTER = "geneFilter", "A subset of genes to include after filtering."
    GENE_HIGHLIGHT = "geneHighlight", "A subset of genes to highlight."
    GENE_SELECTION = "geneSelection" "A subset of genes to select."
    GENE_EXPRESSION_COLORMAP = "geneExpressionColormap", "The colormap to use for the gene expression scale."
    GENE_EXPRESSION_COLORMAP_RANGE = "geneExpressionColormapRange", "The range of gene expression values to map."
    CELL_COLOR_ENCODING = "cellColorEncoding", "The color encoding to use for cell entities."
    SPATIAL_LAYERS = "spatialLayers", "Layer definitions for the spatial view."
    GENOMIC_ZOOM_X = "genomicZoomX", "The zoom level of a higlass view, X dimension."
    GENOMIC_ZOOM_Y = "genomicZoomY", "The zoom level of a higlass view, Y dimension."
    GENOMIC_TARGET_X = "genomicTargetX", "The x-coordinate of the center of a higlass view."
    GENOMIC_TARGET_Y = "genomicTargetY", "The y-coordinate of the center of a higlass view."
    ADDITIONAL_CELL_SETS = "additionalCellSets", "User-defined cell sets."

class Component(DocEnum):
    """
    An enum type representing a view type in the visualization layout.
    """
    SCATTERPLOT = "scatterplot", "The scatterplot component can be used for visualization of 2-dimensional embeddings."
    SPATIAL = "spatial", "The spatial component can be used for visualization of cells, molecules, or images in spatial coordinates."
    DESCRIPTION = "description", "The description component can display short informational text about a dataset."
    STATUS = "status", "The status component can display contextual information such as hover states or error messages."
    CELL_SETS = "cellSets", "The cell sets component can be used to view and manipulate hierarchical (or flat) sets of cells, including sets representing cell type clusters."
    HEATMAP = "heatmap", "The heatmap component can be used to view a cell by gene expression matrix."
    LAYER_CONTROLLER = "layerController", "The layer controller can be used to manipulate channel settings of the images rendered by the spatial component."
    GENOMIC_PROFILES = "genomicProfiles", "The higlass component can be used to visualize genome-wide ATAC-seq profiles."
    CELL_SET_SIZES = "cellSetSizes", "The cell set sizes component can display the quantities of cells in selected cell sets."
    GENES = "genes", "The gene list selector."

class DataType(DocEnum):
    """
    An enum type representing the type of data contained in a file.
    """
    CELLS = "cells", "The cells data type."
    CELL_SETS = "cell-sets", "The cell sets data type."
    EXPRESSION_MATRIX = "expression-matrix", "The gene expression matrix data type."
    MOLECULES = "molecules", "The molecules data type."
    NEIGHBORHOODS = "neighborhoods", "The spatial cell neighborhoods data type."
    RASTER = "raster", "The raster (i.e. imaging) data type."
    GENOMIC_PROFILES = "genomic-profiles", "The genomic profiles data type, used by HiGlass 1D quantitative tracks."

class FileType(DocEnum):
    """
    An enum type representing the file format or schema to which a file conforms.
    """
    EXPRESSION_MATRIX_ZARR = "expression-matrix.zarr", "The Zarr-based expression matrix file type."
    CELLS_JSON = "cells.json", "The JSON-based cells file type."
    MOLECULES_JSON = "molecules.json", "The JSON-based molecules file type."
    NEIGHBORHOODS_JSON = "neighborhoods.json", "The JSON-based neighborhoods file type."
    RASTER_JSON = "raster.json", "The JSON-based raster manifest file type."
    CELL_SETS_JSON = "cell-sets.json", "The JSON-based cell sets file type."
    GENOMIC_PROFILES_ZARR = "genomic-profiles.zarr", "The Zarr-based genomic profile (multivec) file type."




from .constants import DataType as dt

def create_obj_routes(obj, port, dataset_uid, obj_i):
	"""
	For a particular data object, simultaneously set up:
	
	* its server routes and their responses, and
	* the corresponding view config dataset file definitions

	:param obj: An object representing a single-cell data analysis result or microscopy image.
	:type obj: anndata.AnnData or loompy.LoomConnection or zarr.hierarchy.Group
	
	:returns: A list of view config file definitions and a list of server routes.
	:rtype: tuple[list[dict], list[starlette.routing.Route]]
	"""
	obj_file_defs = []
	obj_routes = []

	for data_type in dt:
		try:
			dt_obj_file_defs, dt_obj_routes = obj._get_data(data_type, port, dataset_uid, obj_i)
			obj_file_defs += dt_obj_file_defs
			obj_routes += dt_obj_routes
		except NotImplementedError:
			pass

	return obj_file_defs, obj_routes
	


# Widget dependencies
import ipywidgets as widgets
from traitlets import Unicode, Dict, Int
import time

# Server dependencies
import asyncio
from hypercorn.config import Config
from hypercorn.asyncio import serve

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from threading import Thread
import socket

# Config creation dependencies
from .routes import create_obj_routes

# See js/lib/widget.js for the frontend counterpart to this file.

MAX_PORT_TRIES = 1000

def run_server_loop(app, port):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    config = Config()
    config.bind = [f"localhost:{port}"]

    # As of Hypercorn 0.11.0, need to explicitly set signal handlers to a no-op
    # (otherwise it will try to set signal handlers assuming it is on the main thread which throws an error)
    loop.run_until_complete(serve(app, config, shutdown_trigger=lambda: asyncio.Future()))
    loop.close()

def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0

@widgets.register
class VitessceWidget(widgets.DOMWidget):
    """
    A class to represent a Jupyter widget for Vitessce.
    """

    # Name of the widget view class in front-end
    _view_name = Unicode('VitessceView').tag(sync=True)

    # Name of the widget model class in front-end
    _model_name = Unicode('VitessceModel').tag(sync=True)

    # Name of the front-end module containing widget view
    _view_module = Unicode('vitessce-jupyter').tag(sync=True)

    # Name of the front-end module containing widget model
    _model_module = Unicode('vitessce-jupyter').tag(sync=True)

    # Version of the front-end module containing widget view
    _view_module_version = Unicode('^0.1.0').tag(sync=True)
    # Version of the front-end module containing widget model
    _model_module_version = Unicode('^0.1.0').tag(sync=True)
    
    # Widget specific property.
    # Widget properties are defined as traitlets. Any property tagged with `sync=True`
    # is automatically synced to the frontend *any* time it changes in Python.
    # It is synced back to Python from the frontend *any* time the model is touched.
    config = Dict({}).tag(sync=True)
    height = Int(600).tag(sync=True)
    theme = Unicode('auto').tag(sync=True)

    next_port = 8000

    def __init__(self, config, height=600, theme='auto', port=None):
        """
        Construct a new Vitessce widget.

        :param config: A view config instance.
        :type config: VitessceConfig
        :param str theme: The theme name, either "light" or "dark". By default, "auto", which selects light or dark based on operating system preferences.
        :param int height: The height of the widget, in pixels. By default, 600.
        :param int port: The port to use when serving data objects on localhost. By default, 8000.

        .. code-block:: python
            :emphasize-lines: 4

            from vitessce import VitessceConfig, VitessceWidget

            vc = VitessceConfig.from_object(my_scanpy_object)
            vw = vc.widget()
            vw
        """

        if port is None:
            use_port = VitessceWidget.next_port
            VitessceWidget.next_port += 1
            port_tries = 1
            while is_port_in_use(use_port) and port_tries < MAX_PORT_TRIES:
                use_port = VitessceWidget.next_port
                VitessceWidget.next_port += 1
                port_tries += 1
        else:
            use_port = port
        
        routes = []
        def on_obj(obj, dataset_uid, obj_i):
            obj_file_defs, obj_routes = create_obj_routes(obj, use_port, dataset_uid, obj_i)
            for obj_route in obj_routes:
                routes.append(obj_route)
            return obj_file_defs
        config_dict = config.to_dict(on_obj=on_obj)

        super(VitessceWidget, self).__init__(config=config_dict, height=height, theme=theme)
        
        if len(routes) > 0:
            middleware = [
                Middleware(CORSMiddleware, allow_origins=['*'])
            ]
            app = Starlette(debug=True, routes=routes, middleware=middleware)
            
            t = Thread(target=run_server_loop, args=(app, use_port))
            t.start()
            time.sleep(1)
            
    def _get_coordination_value(self, coordination_type, coordination_scope):
        obj = self.config['coordinationSpace'][coordination_type]
        obj_scopes = list(obj.keys())
        if coordination_scope != None:
            if coordination_scope in obj_scopes:
                return obj[coordination_scope]
            else:
                raise ValueError(f"The specified coordination scope '{coordination_scope}' could not be found for the coordination type '{coordination_type}'. Known coordination scopes are {obj_scopes}")
        else:
            if len(obj_scopes) == 1:
                auto_coordination_scope = obj_scopes[0]
                return obj[auto_coordination_scope]
            elif len(obj_scopes) > 1:
                raise ValueError(f"The coordination scope could not be automatically determined because multiple coordination scopes exist for the coordination type '{coordination_type}'. Please specify one of {obj_scopes} using the scope parameter.")
            else:
                raise ValueError(f"No coordination scopes were found for the coordination type '{coordination_type}'.")
    
    def get_cell_selection(self, scope=None):
        return self._get_coordination_value('cellSelection', scope)




import unittest
from os.path import join

import zarr
from anndata import read_h5ad
from scipy.io import mmread
import pandas as pd
import numpy as np

from create_test_data import (
    create_test_anndata_file,
    create_test_loom_file,
    create_test_ometiff_file,
    create_test_omezarr_store,
    create_test_snaptools_files,
)

from vitessce import (
    OmeTiffWrapper,
    OmeZarrWrapper,
    AnnDataWrapper,
    LoomWrapper,
    SnapWrapper,
)

class TestWrappers(unittest.TestCase):

    def setUp(self):
        create_test_anndata_file(join('data', 'test.h5ad'))
        create_test_loom_file(join('data', 'test.loom'))
        create_test_ometiff_file(join('data', 'test.ome.tif'))
        create_test_omezarr_store(join('data', 'test.ome.zarr'))
        create_test_snaptools_files(
            join('data', 'test.snap.mtx'),
            join('data', 'test.snap.bins.txt'),
            join('data', 'test.snap.barcodes.txt'),
            join('data', 'test.snap.clusters.csv'),
        )

    def test_ome_tiff(self):
        w = OmeTiffWrapper("data/test.ome.tif", offsets_path="data/offsets.json", name="Test")

        raster_json = w.create_raster_json(
            "http://localhost:8000/raster_img",
            "http://localhost:8000/raster_offsets/offsets.json"
        )

        self.assertEqual(raster_json, {
            'images': [
                {
                    'metadata': {
                        'omeTiffOffsetsUrl': 'http://localhost:8000/raster_offsets/offsets.json'
                    },
                    'name': 'Test',
                    'type': 'ome-tiff',
                    'url': 'http://localhost:8000/raster_img'
                }
            ],
            'schemaVersion': '0.0.2'
        })

        obj_file_defs, obj_routes = w.get_raster(8000, 'A', 0)

        self.assertEqual(obj_file_defs, [
            {
                'fileType': 'raster.json',
                'type': 'raster',
                'url': 'http://localhost:8000/A/0/raster'
            }
        ])
    
    def test_omezarr_store(self):
        z = zarr.open('data/test.ome.zarr')
        w = OmeZarrWrapper(z)

        raster_json = w.create_raster_json(
            "http://localhost:8000/raster_img"
        )
        
        # TODO
        # self.assertEqual(raster_json, {})

        obj_file_defs, obj_routes = w.get_raster(8000, 'A', 0)
        self.assertEqual(obj_file_defs, [
            {
                'fileType': 'raster.json',
                'type': 'raster',
                'url': 'http://localhost:8000/A/0/raster'
            }
        ])
    
    def test_base_url(self):
        z = zarr.open('data/test.ome.zarr')
        w = OmeZarrWrapper(z, base_url="https://example.com")

        raster_json = w.create_raster_json(
            "https://example.com/raster_img"
        )
        
        # TODO
        # self.assertEqual(raster_json, {})

        obj_file_defs, obj_routes = w.get_raster(8000, 'A', 0)
        self.assertEqual(obj_file_defs, [
            {
                'fileType': 'raster.json',
                'type': 'raster',
                'url': 'https://example.com/A/0/raster'
            }
        ])
    
    def test_anndata(self):
        adata = read_h5ad(join('data', 'test.h5ad'))
        w = AnnDataWrapper(adata, cell_set_obs_cols=['CellType'])

        cells_json = w.create_cells_json()
        cell_sets_json = w.create_cell_sets_json()

        obj_file_defs, obj_routes = w.get_cells(8000, 'A', 0)
        self.assertEqual(obj_file_defs, [{'type': 'cells', 'fileType': 'cells.json', 'url': 'http://localhost:8000/A/0/cells'}])

        obj_file_defs, obj_routes = w.get_cell_sets(8000, 'A', 0)
        self.assertEqual(obj_file_defs, [{'type': 'cell-sets', 'fileType': 'cell-sets.json', 'url': 'http://localhost:8000/A/0/cell-sets'}])

    def test_snaptools(self):
        mtx = mmread(join('data', 'test.snap.mtx'))
        barcodes_df = pd.read_csv(join('data', 'test.snap.barcodes.txt'), header=None)
        bins_df = pd.read_csv(join('data', 'test.snap.bins.txt'), header=None)
        clusters_df = pd.read_csv(join('data', 'test.snap.clusters.csv'), index_col=0)

        zarr_filepath = join('data', 'test_out.snap.multivec.zarr')

        w = SnapWrapper(mtx, barcodes_df, bins_df, clusters_df)
        w.create_genomic_multivec_zarr(zarr_filepath)

        z = zarr.open(zarr_filepath, mode='r')

        self.assertEqual(z['chromosomes/chr1/5000'].shape, (4, 49792))
        self.assertEqual(z['chromosomes/chr1/5000'][:,0].sum(), 0)
        self.assertEqual(z['chromosomes/chr1/5000'][:,1].sum(), 0)
        self.assertEqual(z['chromosomes/chr1/5000'][:,2].sum(), 7)
        self.assertEqual(z['chromosomes/chr1/5000'][:,3].sum(), 7)
        self.assertEqual(z['chromosomes/chr1/5000'][:,4].sum(), 0)
        self.assertEqual(z['chromosomes/chr1/5000'][0,:].sum(), 17)
        self.assertEqual(z['chromosomes/chr1/10000'][0,:].sum(), 17)
        self.assertEqual(z['chromosomes/chr1/5000'][:,2].sum(), 7)
        self.assertEqual(z['chromosomes/chr2/5000'][:,0].sum(), 0)
        self.assertEqual(z['chromosomes/chr2/5000'][:,1].sum(), 0)
        self.assertEqual(z['chromosomes/chr2/10000'][:,0].sum(), 0)
        self.assertEqual(z['chromosomes/chr2/5000'][:,2].sum(), 4)
        self.assertEqual(z['chromosomes/chr2/5000'][:,3].sum(), 9)
        self.assertEqual(z['chromosomes/chr2/10000'][:,1].sum(), 13)
        self.assertEqual(z['chromosomes/chr3/5000'][:,3].sum(), 9)
        self.assertEqual(z['chromosomes/chr3/5000'][:].sum(), 9)
        self.assertEqual(z['chromosomes/chr18/5000'][:].sum(), 8)

        cells_json = w.create_cells_json()
        self.assertEqual(len(cells_json), 6)
        self.assertEqual(cells_json['AAACATCGAGTACAAGACAGCAGA'], { 'mappings': { 'UMAP': [4.43, 1.64] } })
        


import json
from uuid import uuid4

from .constants import (
    CoordinationType as ct,
    Component as cm,
    DataType as dt,
    FileType as ft
)

from .wrappers import (
    AnnDataWrapper,
    SnapWrapper,
)
from .widget import VitessceWidget

def _get_next_scope(prev_scopes):
    chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    next_char_indices = [0]

    def next():
        r = []
        for char_index in next_char_indices:
            r = [chars[char_index]] + r
        increment = True
        for i in range(len(next_char_indices)):
            next_char_indices[i] += 1
            val = next_char_indices[i]
            if val >= len(chars):
                next_char_indices[i] = 0
            else:
                increment = False
                break
        
        if increment:
            next_char_indices.append(0)
        
        return "".join([ str(j) for j in r ])
    
    next_scope = next()
    while next_scope in prev_scopes:
        next_scope = next()
    
    return next_scope

class VitessceConfigDatasetFile:
    """
    A class to represent a file (described by a URL, data type, and file type) in a Vitessce view config dataset.
    """
    def __init__(self, url, data_type, file_type):
        """
        Not meant to be instantiated directly, but instead created and returned by the ``VitessceConfigDataset.add_file()`` method.

        :param str url: A URL to this file. Can be a localhost URL or a remote URL.
        :param str data_type: A data type.
        :param str file_type: A file type.
        """
        self.file = {
            "url": url,
            "type": data_type,
            "fileType": file_type,
        }
    
    def to_dict(self):
        return self.file

class VitessceConfigDataset:
    """
    A class to represent a dataset (i.e. list of files containing common biological entities) in the Vitessce view config.
    """
    def __init__(self, uid, name):
        """
        Not meant to be instantiated directly, but instead created and returned by the ``VitessceConfig.add_dataset()`` method.

        :param str uid: A unique identifier for this dataset.
        :param str name: A name for this dataset.
        """
        self.dataset = {
            "uid": uid,
            "name": name,
            "files": [],
        }
        self.objs = []
    
    def add_file(self, url, data_type, file_type):
        """
        Add a new file definition to this dataset instance.

        :param str url: The URL for the file, pointing to either a local or remote location.
        :param data_type: The type of data stored in the file. Must be compatible with the specified file type.
        :type data_type: str or vitessce.constants.DataType
        :param file_type: The file type. Must be compatible with the specified data type.
        :type file_type: str or vitessce.constants.FileType

        :returns: Self, to allow function chaining.
        :rtype: VitessceConfigDataset

        .. code-block:: python
            :emphasize-lines: 6-10

            from vitessce import VitessceConfig, DataType as dt, FileType as ft

            vc = VitessceConfig(name='My Config')
            my_dataset = (
                vc.add_dataset(name='My Dataset')
                .add_file(
                    url="http://example.com/cells.json",
                    data_type=dt.CELLS,
                    file_type=ft.CELLS_JSON,
                )
            )
        """

        assert type(data_type) == str or type(data_type) == dt
        assert type(file_type) == str or type(file_type) == ft

        # TODO: assert that the file type is compatible with the data type (and vice versa)

        if type(data_type) == str:
            data_type_str = data_type
        else:
            data_type_str = data_type.value
        
        if type(file_type) == str:
            file_type_str = file_type
        else:
            file_type_str = file_type.value

        self.dataset["files"].append(VitessceConfigDatasetFile(url=url, data_type=data_type_str, file_type=file_type_str))
        return self
    
    def add_object(self, obj):
        """
        Add a data object to this dataset instance.

        :param obj: A data object that can be served locally or uploaded to a remote storage provider.
        :type obj: anndata.AnnData or loompy.LoomConnection or zarr.Store

        :returns: Self, to allow function chaining.
        :rtype: VitessceConfigDataset
        """
        self.objs.append(obj)
        return self

    def to_dict(self, on_obj):
        obj_file_defs = []
        for obj_i, obj in enumerate(self.objs):
            if on_obj is not None:
                obj_file_defs += on_obj(obj, self.dataset["uid"], obj_i)
                
        return {
            **self.dataset,
            "files": [ f.to_dict() for f in self.dataset["files"] ] + obj_file_defs,
        }

class VitessceConfigViewHConcat:
    """
    A class to represent a horizontal concatenation of view instances.
    """
    def __init__(self, views):
        """
        Not meant to be instantiated directly, but instead created and returned by the ``hconcat`` helper function.
        """
        self.views = views
    
    def __or__(self, other):
        return hconcat(self, other)
    
    def __truediv__(self, other):
        return vconcat(self, other)

def hconcat(*views):
    """
    A helper function to create a ``VitessceConfigViewHConcat`` instance.

    :param \*views: A variable number of views to concatenate horizontally.
    :type \*views: VitessceConfigView or VitessceConfigViewVConcat or VitessceConfigViewHConcat

    :returns: The concatenated view instance.
    :rtype: VitessceConfigViewHConcat

    .. code-block:: python
        :emphasize-lines: 8
        
        from vitessce import VitessceConfig, Component as cm, hconcat, vconcat

        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')
        v1 = vc.add_view(my_dataset, cm.SPATIAL)
        v2 = vc.add_view(my_dataset, cm.SPATIAL)
        v3 = vc.add_view(my_dataset, cm.SPATIAL)
        vc.layout(hconcat(v1, vconcat(v2, v3)))
    """
    return VitessceConfigViewHConcat(views)

class VitessceConfigViewVConcat:
    """
    A class to represent a vertical concatenation of view instances.
    """
    def __init__(self, views):
        """
        Not meant to be instantiated directly, but instead created and returned by the ``vconcat`` helper function.
        """
        self.views = views
    
    def __or__(self, other):
        return hconcat(self, other)
    
    def __truediv__(self, other):
        return vconcat(self, other)

def vconcat(*views):
    """
    A helper function to create a ``VitessceConfigViewVConcat`` instance.

    :param \*views: A variable number of views to concatenate vertically.
    :type \*views: VitessceConfigView or VitessceConfigViewVConcat or VitessceConfigViewHConcat

    :returns: The concatenated view instance.
    :rtype: VitessceConfigViewVConcat

    .. code-block:: python
        :emphasize-lines: 8
        
        from vitessce import VitessceConfig, Component as cm, hconcat, vconcat

        vc = VitessceConfig()
        my_dataset = vc.add_dataset(name='My Dataset')
        v1 = vc.add_view(my_dataset, cm.SPATIAL)
        v2 = vc.add_view(my_dataset, cm.SPATIAL)
        v3 = vc.add_view(my_dataset, cm.SPATIAL)
        vc.layout(hconcat(v1, vconcat(v2, v3)))
    """
    return VitessceConfigViewVConcat(views)

class VitessceConfigView:
    """
    A class to represent a view (i.e. visualization component) in the Vitessce view config layout.
    """
    def __init__(self, component, coordination_scopes, x, y, w, h):
        """
        Not meant to be instantiated directly, but instead created and returned by the ``VitessceConfig.add_view()`` method.

        :param str component: The name of the component used for this view.
        :param dict coordination_scopes: A mapping of coordination types to coordination scopes.
        :param int x: An x-coordinate for this view in the grid.
        :param int y: A y-coordinate for this view in the grid.
        :param int w: A width for this view in the grid.
        :param int h: A height for this view in the grid.
        """
        self.view = {
            "component": component,
            "coordinationScopes": coordination_scopes,
            "x": x,
            "y": y,
            "w": w,
            "h": h
        }
    
    def use_coordination(self, *c_scopes):
        """
        Attach a coordination scope to this view instance. All views using the same coordination scope for a particular coordination type will effectively be linked together.

        :param \*c_scopes: A variable number of coordination scope instances can be passed.
        :type \*c_scopes: VitessceConfigCoordinationScope

        :returns: Self, to allow chaining.
        :rtype: VitessceConfigView

        .. code-block:: python
            :emphasize-lines: 12-13

            from vitessce import VitessceConfig, Component as cm, CoordinationType as ct

            vc = VitessceConfig()
            my_dataset = vc.add_dataset(name='My Dataset')
            v1 = vc.add_view(my_dataset, cm.SPATIAL)
            v2 = vc.add_view(my_dataset, cm.SPATIAL)
            zoom_scope, x_scope, y_scope = vc.add_coordination(
                ct.SPATIAL_ZOOM,
                ct.SPATIAL_TARGET_X,
                ct.SPATIAL_TARGET_Y,
            )
            v1.use_coordination(zoom_scope, x_scope, y_scope)
            v2.use_coordination(zoom_scope, x_scope, y_scope)
            zoom_scope.set_value(2)
            x_scope.set_value(0)
            y_scope.set_value(0)
        """
        for c_scope in c_scopes:
            assert type(c_scope) == VitessceConfigCoordinationScope
            self.view["coordinationScopes"][c_scope.c_type] = c_scope.c_scope
        return self
    
    def _set_xywh(self, x, y, w, h):
        self.view["x"] = x
        self.view["y"] = y
        self.view["w"] = w
        self.view["h"] = h
    
    def set_props(self, **kwargs):
        if "props" in self.view.keys():
            self.view["props"] = {
                **self.view["props"],
                **kwargs
            }
        else:
            self.view["props"] = kwargs
    
    def to_dict(self):
        return self.view

    def __or__(self, other):
        return hconcat(self, other)
    
    def __truediv__(self, other):
        return vconcat(self, other)


class VitessceConfigCoordinationScope:
    """
    A class to represent a coordination scope in the Vitessce view config coordination space.
    """
    def __init__(self, c_type, c_scope):
        """
        Not meant to be instantiated directly, but instead created and returned by the ``VitessceConfig.add_coordination()`` method.

        :param str c_type: The coordination type for this coordination scope.
        :param str c_scope: The coordination scope name.
        """
        self.c_type = c_type
        self.c_scope = c_scope
        self.c_value = None
    
    def set_value(self, c_value):
        """
        Set the value of the coordination scope.

        :param any c_value: The coordination value to be set. Can be any value that is valid for the coordination type. Must be serializable to JSON.

        :returns: Self, to allow chaining.
        :rtype: VitessceConfigCoordinationScope

        .. code-block:: python
            :emphasize-lines: 14-16

            from vitessce import VitessceConfig, Component as cm, CoordinationType as ct

            vc = VitessceConfig()
            my_dataset = vc.add_dataset(name='My Dataset')
            v1 = vc.add_view(my_dataset, cm.SPATIAL)
            v2 = vc.add_view(my_dataset, cm.SPATIAL)
            zoom_scope, x_scope, y_scope = vc.add_coordination(
                ct.SPATIAL_ZOOM,
                ct.SPATIAL_TARGET_X,
                ct.SPATIAL_TARGET_Y,
            )
            v1.use_coordination(zoom_scope, x_scope, y_scope)
            v2.use_coordination(zoom_scope, x_scope, y_scope)
            zoom_scope.set_value(2)
            x_scope.set_value(0)
            y_scope.set_value(0)
        """
        self.c_value = c_value
        return self

class VitessceConfig:
    """
    A class to represent a Vitessce view config.
    """

    def __init__(self, name=None, description=None):
        """
        Construct a Vitessce view config object.

        :param str name: A name for the view config. Optional.
        :param str description: A description for the view config. Optional.

        .. code-block:: python
            :emphasize-lines: 3
            
            from vitessce import VitessceConfig

            vc = VitessceConfig(name='My Config')
        """
        self.config = {
            "version": "1.0.0",
            "name": name,
            "description": description,
            "datasets": [],
            "coordinationSpace": {},
            "layout": [],
            "initStrategy": "auto"
        }

        if name is None:
            self.config["name"] = ""
        else:
            self.config["name"] = name
        if description is None:
            self.config["description"] = ""
        else:
            self.config["description"] = description


    def add_dataset(self, name="", uid=None):
        """
        Add a dataset to the config.

        :param str name: A name for this dataset.
        :param str uid: A unique identifier for this dataset. Optional. If None, will be automatically generated.

        :returns: The instance for the new dataset.
        :rtype: VitessceConfigDataset
        
        .. code-block:: python
            :emphasize-lines: 5

            from vitessce import VitessceConfig, DataType as dt, FileType as ft

            vc = VitessceConfig(name='My Config')
            my_dataset = (
                vc.add_dataset(name='My Dataset')
                .add_file(
                    url="http://example.com/cells.json",
                    data_type=dt.CELLS,
                    file_type=ft.CELLS_JSON,
                )
            )
        """
        uid = uid if uid is not None else _get_next_scope([ d.dataset['uid'] for d in self.config["datasets"] ])
        assert type(uid) == str
        vcd = VitessceConfigDataset(uid, name)
        self.config["datasets"].append(vcd)
        [d_scope] = self.add_coordination(ct.DATASET)
        d_scope.set_value(uid)
        return vcd
    
    def add_view(self, dataset, component, x=0, y=0, w=1, h=1, mapping=None):
        """
        Add a view to the config.

        :param dataset: A dataset instance to be used for the data visualized in this view.
        :type dataset: VitessceConfigDataset
        :param component: A component name, either as a string or using the Component enum values.
        :type component: str or vitessce.constants.Component

        :param str mapping: An optional convenience parameter for setting the EMBEDDING_TYPE coordination scope value. This parameter is only applicable to the SCATTERPLOT component.
        :param int x: The horizontal position of the view. Must be an integer between 0 and 11. Optional.
        :param int y: The vertical position of the view. Must be an integer between 0 and 11. Optional.
        :param int w: The width of the view. Must be an integer between 1 and 12. Optional.
        :param int h: The height of the view. Must be an integer between 1 and 12. Optional.

        :returns: The instance for the new view.
        :rtype: VitessceConfigView

        .. code-block:: python
            :emphasize-lines: 5-6

            from vitessce import VitessceConfig, Component as cm

            vc = VitessceConfig()
            my_dataset = vc.add_dataset(name='My Dataset')
            v1 = vc.add_view(my_dataset, cm.SPATIAL)
            v2 = vc.add_view(my_dataset, cm.SCATTERPLOT, mapping="X_umap")
        """
        assert type(dataset) == VitessceConfigDataset
        assert type(component) == str or type(component) == cm

        if type(component) == str:
            component_str = component
        else:
            component_str = component.value

        # Find the coordination scope name associated with the dataset
        dataset_matches = [
            scope_name
            for scope_name, dataset_scope in self.config["coordinationSpace"][ct.DATASET.value].items()
            if dataset_scope.c_value == dataset.dataset["uid"]
        ] if ct.DATASET.value in self.config["coordinationSpace"].keys() else []
        if len(dataset_matches) == 1:
            dataset_scope = dataset_matches[0]
        else:
            raise ValueError("No coordination scope matching the dataset parameter could be found in the coordination space.")

        # Set up the view's dataset coordination scope based on the dataset parameter.
        coordination_scopes = {
            ct.DATASET.value: dataset_scope,
        }
        vcv = VitessceConfigView(component_str, coordination_scopes, x, y, w, h)
        
        # Use the mapping parameter if component is scatterplot and the mapping is not None
        if mapping is not None:
            [et_scope] = self.add_coordination(ct.EMBEDDING_TYPE)
            et_scope.set_value(mapping)
            vcv.use_coordination(et_scope)
        self.config["layout"].append(vcv)
        return vcv
    

    def add_coordination(self, *c_types):
        """
        Add scope(s) for new coordination type(s) to the config.

        :param \*c_types: A variable number of coordination types.
        :type \*c_types: str or vitessce.constants.CoordinationType
        
        :returns: The instances for the new scope objects corresponding to each coordination type. These can be linked to views via the ``VitessceConfigView.use_coordination()`` method.
        :rtype: list[VitessceConfigCoordinationScope]

        .. code-block:: python
            :emphasize-lines: 7-11

            from vitessce import VitessceConfig, Component as cm, CoordinationType as ct

            vc = VitessceConfig()
            my_dataset = vc.add_dataset(name='My Dataset')
            v1 = vc.add_view(my_dataset, cm.SPATIAL)
            v2 = vc.add_view(my_dataset, cm.SPATIAL)
            zoom_scope, x_scope, y_scope = vc.add_coordination(
                ct.SPATIAL_ZOOM,
                ct.SPATIAL_TARGET_X,
                ct.SPATIAL_TARGET_Y,
            )
            v1.use_coordination(zoom_scope, x_scope, y_scope)
            v2.use_coordination(zoom_scope, x_scope, y_scope)
            zoom_scope.set_value(2)
            x_scope.set_value(0)
            y_scope.set_value(0)
        """
        result = []
        for c_type in c_types:
            assert type(c_type) == ct or type(c_type) == str
            if type(c_type) == ct:
                c_type_str = c_type.value
            else:
                c_type_str = c_type
            prev_scopes = list(self.config["coordinationSpace"][c_type_str].keys()) if c_type_str in self.config["coordinationSpace"].keys() else []
            scope = VitessceConfigCoordinationScope(c_type_str, _get_next_scope(prev_scopes))
            if scope.c_type not in self.config["coordinationSpace"]:
                self.config["coordinationSpace"][scope.c_type] = {}
            self.config["coordinationSpace"][scope.c_type][scope.c_scope] = scope
            result.append(scope)
        return result
    
    def link_views(self, views, c_types, c_values = None):
        """
        A convenience function for setting up new coordination scopes across a set of views.
        
        :param views: views An array of view objects to link together.
        :type views: list of VitessceConfigView
        :param c_types: The coordination types on which to coordinate the views.
        :type c_types: list of str or list of vitessce.constants.CoordinationType
        :param list c_values: Initial values corresponding to each coordination type. Should have the same length as the c_types array. Optional.

        :returns: Self, to allow chaining.
        :rtype: VitessceConfig
        """
        c_scopes = self.add_coordination(*c_types)
        for view in views:
            for c_scope in c_scopes:
                view.use_coordination(c_scope)

        if c_values is not None and len(c_values) == len(c_types):
            for i, c_scope in enumerate(c_scopes):
                c_scope.set_value(c_values[i])

        return self
    
    def layout(self, view_concat):
        """
        Create a multi-view layout based on (potentially recursive) view concatenations.

        :param view_concat: Views arranged by concatenating vertically or horizontally. Alternatively, a single view can be passed.
        :type view_concat: VitessceConfigViewHConcat or VitessceConfigViewVConcat or VitessceConfigView

        :returns: Self, to allow chaining.
        :rtype: VitessceConfig
        
        .. code-block:: python
            :emphasize-lines: 8
            
            from vitessce import VitessceConfig, Component as cm, hconcat, vconcat

            vc = VitessceConfig()
            my_dataset = vc.add_dataset(name='My Dataset')
            v1 = vc.add_view(my_dataset, cm.SPATIAL)
            v2 = vc.add_view(my_dataset, cm.SPATIAL)
            v3 = vc.add_view(my_dataset, cm.SPATIAL)
            vc.layout(hconcat(v1, vconcat(v2, v3)))
        
        .. code-block:: python
            :emphasize-lines: 8
            
            from vitessce import VitessceConfig, Component as cm

            vc = VitessceConfig()
            my_dataset = vc.add_dataset(name='My Dataset')
            v1 = vc.add_view(my_dataset, cm.SPATIAL)
            v2 = vc.add_view(my_dataset, cm.SPATIAL)
            v3 = vc.add_view(my_dataset, cm.SPATIAL)
            vc.layout(v1 | (v2 / v3)) # * magic * (alternative syntax)
        """

        def _layout(obj, x_min, x_max, y_min, y_max):
            w = x_max - x_min
            h = y_max - y_min
            if type(obj) == VitessceConfigView:
                obj._set_xywh(x_min, y_min, w, h)
            elif type(obj) == VitessceConfigViewHConcat:
                views = obj.views
                num_views = len(views)
                for i in range(num_views):
                    _layout(
                        views[i],
                        x_min+(w/num_views)*i,
                        x_min+(w/num_views)*(i+1),
                        y_min,
                        y_max
                    )
            elif type(obj) == VitessceConfigViewVConcat:
                views = obj.views
                num_views = len(views)
                for i in range(num_views):
                    _layout(
                        views[i],
                        x_min,
                        x_max,
                        y_min+(h/num_views)*i,
                        y_min+(h/num_views)*(i+1),
                    )

        # Recursively set the values (x,y,w,h) for each view.
        _layout(view_concat, 0, 12, 0, 12)

        # TODO: decide how to handle views that were omitted from the `view_concat` parameter
        # TODO: decide how to handle .add_view() being called after .layout() has been called

        return self
        
    def to_dict(self, on_obj=None):
        """
        Convert the view config instance to a dict object.

        :param on_obj: This callback is required only if datasets within the view config contain objects added via the ``VitessceConfigDataset.add_object`` method (rather than only files added via the ``VitessceConfigDataset.add_file`` method). This function must take the data object as a parameter, and return a list of valid file definition dicts (URL, data type, file type). This parameter is primarily intended to be used internally by the ``VitessceWidget`` class.
        :type on_obj: function or None

        :returns: The view config as a dict. Useful for serializing to JSON.
        :rtype: dict
        """
        return {
            **self.config,
            "datasets": [ d.to_dict(on_obj) for d in self.config["datasets"] ],
            "coordinationSpace": dict([
                (c_type, dict([
                    (c_scope_name, c_scope.c_value) for c_scope_name, c_scope in c_scopes.items() 
                ])) for c_type, c_scopes in self.config["coordinationSpace"].items()
            ]),
            # TODO: compute the x,y,w,h values if not explicitly defined
            "layout": [ c.to_dict() for c in self.config["layout"] ]
        }

    @staticmethod
    def from_dict(config):
        """
        Helper function to construct a Vitessce view config object from an existing config.

        :param dict config: An existing Vitessce view config as a dict to allow manipulation through the ``VitessceConfig`` API.
        
        :returns: The config instance.
        :rtype: VitessceConfig

        .. code-block:: python
            :emphasize-lines: 3
            
            from vitessce import VitessceConfig

            vc = VitessceConfig.from_dict(my_existing_config)
        """
        # TODO: Validate the incoming config.

        vc = VitessceConfig(name=config["name"], description=config["description"])

        # Add each dataset from the incoming config.
        for d in config["datasets"]:
            new_dataset = vc.add_dataset(uid=d["uid"], name=d["name"])
            for f in d["files"]:
                new_file = new_dataset.add_file(
                    url=f["url"],
                    data_type=f["type"],
                    file_type=f["fileType"]
                )
        
        for c_type in config['coordinationSpace'].keys():
            if c_type != ct.DATASET.value:
                c_obj = config['coordinationSpace'][c_type]
                vc.config['coordinationSpace'][c_type] = {}
                for c_scope_name, c_scope_value in c_obj.items():
                    scope = VitessceConfigCoordinationScope(c_type, c_scope_name)
                    scope.set_value(c_scope_value)
                    vc.config['coordinationSpace'][c_type][c_scope_name] = scope
        
        for c in config['layout']:
            c_coord_scopes = c['coordinationScopes'] if 'coordinationScopes' in c.keys() else {}
            new_view = VitessceConfigView(c['component'], c_coord_scopes, c['x'], c['y'], c['w'], c['h'])
            if 'props' in c.keys():
                new_view.set_props(**c['props'])
            vc.config['layout'].append(new_view)

        return vc


    @staticmethod
    def from_object(obj, name=None, description=None):
        """
        Helper function to automatically construct a Vitessce view config object from a single-cell dataset object.
        Particularly helpful when using the ``VitessceWidget`` Jupyter widget.

        :param obj: A single-cell dataset in the format of a commonly-used single-cell or imaging data analysis package.
        :type obj: anndata.AnnData or loompy.LoomConnection or zarr.Store
        
        :returns: The config instance.
        :rtype: VitessceConfig

        .. code-block:: python
            :emphasize-lines: 3
            
            from vitessce import VitessceConfig
            
            vc = VitessceConfig.from_object(my_scanpy_object)
        """
        vc = VitessceConfig(name=name, description=description)
        dataset = vc.add_dataset()

        if type(obj) == AnnDataWrapper:
            dataset.add_object(obj)

            # TODO: use the available embeddings to determine how many / which scatterplots to add.
            scatterplot = vc.add_view(dataset, cm.SCATTERPLOT, mapping="X_umap")
            cell_sets = vc.add_view(dataset, cm.CELL_SETS)
            genes = vc.add_view(dataset, cm.GENES)
            heatmap = vc.add_view(dataset, cm.HEATMAP)

            vc.layout((scatterplot | (cell_sets / genes)) / heatmap)
        
        elif type(obj) == SnapWrapper:
            dataset.add_object(obj)

            genomic_profiles = vc.add_view(dataset, cm.GENOMIC_PROFILES)
            scatter = vc.add_view(dataset, cm.SCATTERPLOT, mapping = "UMAP")
            cell_sets = vc.add_view(dataset, cm.CELL_SETS)

            vc.layout(genomic_profiles / (scatter | cell_sets))
        
        else:
            print("Encountered an unknown object type. Unable to automatically generate a Vitessce view config.")

        return vc
    
    def widget(self, **kwargs):
        """
        Convenience function for instantiating a VitessceWidget object based on this config.
        
        :returns: The Jupyter widget.
        :rtype: VitessceConfig

        .. code-block:: python
            :emphasize-lines: 6-7

            from vitessce import VitessceConfig, Component as cm, CoordinationType as ct

            vc = VitessceConfig()
            my_dataset = vc.add_dataset(name='My Dataset')
            v1 = vc.add_view(my_dataset, cm.SPATIAL)
            vc.layout(v1)
            vw = vc.widget()
            vw
        """
        return VitessceWidget(self, **kwargs)


